from .lib.attrdict import AttrDict

class HTTPResponse:	
	def __init__ (self, handler):
		self.handler = handler
		self.response = handler.response
		self.request = self.response.request
	
	def __del__ (self):
		self.response.request = None
		self.request = None		
		self.handler.response = None
		self.response = None
		self.handler = None
			
	def __getattr__ (self, name):
		return getattr (self.response, name)



import winreg
from sys import exit
import time 
import win32api
import win32con
from . import win32service
wa, wc, ws = win32api, win32con, win32service

class WService:
	"""
	The WService Class is used for controlling WinNT, Win2k & WinXP like
	services. Just pass the name of the service you wish to control to the
	class instance and go from there. For example, if you want to control 
	the Workstation service try this:
	 
		import WService
		workstation = WService.WService("Workstation")
		workstation.start()
		workstation.fetchstatus("running", 10)
		workstation.stop()
		workstation.fetchstatus("stopped")

	Creating an instance of the WService class is done by passing the name of
	the service as it appears in the Management Console or the short name as
	it appears in the registry. Mixed case is ok.
		cvs = WService.WService("CVS NT Service 1.11.1.2 (Build 41)")		
			or
		cvs = WService.WService("cvs")

	If needing remote service control try this:
		cvs = WService.WService("cvs", r"\\CVS_SERVER")
			or
		cvs = WService.WService("cvs", "\\\\CVS_SERVER")  
		
	The WService Class supports these methods:
		
		start:		  Starts service.
		stop:		   Stops service.
		restart:		Stops and restarts service.
		pause:		  Pauses service (Only if service supports feature).
		resume:		 Resumes service that has been paused.
		status:		 Queries current status of service.
		fetchstatus:	Continually queries service until requested status(STARTING, RUNNING,
							STOPPING & STOPPED) is met or timeout value(in seconds) reached.
							Default timeout value is infinite.							  
		infotype:	   Queries service for process type. (Single, shared and/or 
							interactive process)
		infoctrl:	   Queries control information about a running service.
							i.e. Can it be paused, stopped, etc?
		infostartup:	Queries service Startup type. (Boot, System, 
							Automatic, Manual, Disabled)
		setstartup	  Changes/sets Startup type. (Boot, System, 
							Automatic, Manual, Disabled)					
		getname:		Gets the long and short service names used by Windows.
							(Generally used for internal purposes)
	"""	
		
	def __init__(self, service, machinename=None, dbname=None):
		self.userv = service
		self.scmhandle = ws.OpenSCManager(machinename, dbname, ws.SC_MANAGER_ALL_ACCESS)
		self.sserv, self.lserv = self.getname()
		if (self.sserv or self.lserv) == None: exit()
		self.handle = ws.OpenService(self.scmhandle, self.sserv, ws.SERVICE_ALL_ACCESS)
		self.sccss = "SYSTEM\\CurrentControlSet\\Services\\"
		
	def start(self):
		ws.StartService(self.handle, None)
		
	def stop(self):
		self.stat = ws.ControlService(self.handle, ws.SERVICE_CONTROL_STOP)

	def restart(self):
		self.stop()
		self.fetchstate("STOPPED")
		self.start()
		
	def pause(self):
		self.stat = ws.ControlService(self.handle, ws.SERVICE_CONTROL_PAUSE)
	
	def resume(self):
		self.stat = ws.ControlService(self.handle, ws.SERVICE_CONTROL_CONTINUE)
		
	def status(self, prn = 0):
		self.stat = ws.QueryServiceStatus(self.handle)
		if self.stat[1]==ws.SERVICE_STOPPED:
			if prn == 1:
				print("The", self.lserv, "service is stopped.")
			else:
				return "STOPPED"
		elif self.stat[1]==ws.SERVICE_START_PENDING:
			if prn == 1:
				print("The", self.lserv, "service is starting.")
			else:
				return "STARTING"
		elif self.stat[1]==ws.SERVICE_STOP_PENDING:
			if prn == 1:
				print("The", self.lserv, "service is stopping.")		
			else:
				return "STOPPING"
		elif self.stat[1]==ws.SERVICE_RUNNING:
			if prn == 1:
				print("The", self.lserv, "service is running.")		
			else:
				return "RUNNING"
				
	def fetchstatus(self, fstatus, timeout=None):
		self.fstatus = fstatus.upper()
		if timeout != None:
			timeout = int(timeout); timeout *= 2
		def to(timeout):
			time.sleep(.5)
			if timeout != None:
				if timeout > 1:
					timeout -= 1; return timeout
				else:
					return "TO"   
		if self.fstatus == "STOPPED":
			while 1:
				self.stat = ws.QueryServiceStatus(self.handle)
				if self.stat[1]==ws.SERVICE_STOPPED:
					self.fstate = "STOPPED"; break
				else:
					timeout=to(timeout)
					if timeout == "TO":
						return "TIMEDOUT"; break
		elif self.fstatus == "STOPPING":
			while 1:
				self.stat = ws.QueryServiceStatus(self.handle)
				if self.stat[1]==ws.SERVICE_STOP_PENDING:
					self.fstate = "STOPPING"; break
				else:
					timeout=to(timeout)
					if timeout == "TO":
						return "TIMEDOUT"; break					  
		elif self.fstatus == "RUNNING":
			while 1:
				self.stat = ws.QueryServiceStatus(self.handle)
				if self.stat[1]==ws.SERVICE_RUNNING:
					self.fstate = "RUNNING"; break
				else:
					timeout=to(timeout)
					if timeout == "TO":
						return "TIMEDOUT"; break					   
		elif self.fstatus == "STARTING":
			while 1:
				self.stat = ws.QueryServiceStatus(self.handle)
				if self.stat[1]==ws.SERVICE_START_PENDING:
					self.fstate = "STARTING"; break
				else:
					timeout=to(timeout)
					if timeout == "TO":
						return "TIMEDOUT"; break
		
	def infotype(self):
		self.stat = ws.QueryServiceStatus(self.handle)
		if self.stat[0] & ws.SERVICE_WIN32_OWN_PROCESS:
			print("The", self.lserv, "service runs in its own process.")
		if self.stat[0] & ws.SERVICE_WIN32_SHARE_PROCESS:
			print("The", self.lserv, "service shares a process with other services.")
		if self.stat[0] & ws.SERVICE_INTERACTIVE_PROCESS:
			print("The", self.lserv, "service can interact with the desktop.")
		
	def infoctrl(self):
		self.stat = ws.QueryServiceStatus(self.handle)
		if self.stat[2] & ws.SERVICE_ACCEPT_PAUSE_CONTINUE:
			print("The", self.lserv, "service can be paused.")
		if self.stat[2] & ws.SERVICE_ACCEPT_STOP:
			print("The", self.lserv, "service can be stopped.")
		if self.stat[2] & ws.SERVICE_ACCEPT_SHUTDOWN:
			print("The", self.lserv, "service can be shutdown.")	

	def infostartup(self):
		self.isuphandle = wa.RegOpenKeyEx(wc.HKEY_LOCAL_MACHINE, self.sccss + self.sserv, 0, wc.KEY_READ)
		self.isuptype = wa.RegQueryValueEx(self.isuphandle, "Start")[0]
		wa.RegCloseKey(self.isuphandle)
		if self.isuptype == 0:
			return "boot"
		elif self.isuptype == 1:
			return "system"
		elif self.isuptype == 2:
			return "automatic"
		elif self.isuptype == 3:
			return "manual"
		elif self.isuptype == 4:
			return "disabled"
		 
	def setstartup(self, startuptype):
		self.startuptype = startuptype.lower()
		if self.startuptype == "boot":
			self.suptype = 0
		elif self.startuptype == "system":
			self.suptype = 1
		elif self.startuptype == "automatic":
			self.suptype = 2
		elif self.startuptype == "manual":
			self.suptype = 3
		elif self.startuptype == "disabled":
			self.suptype = 4
		self.snc = ws.SERVICE_NO_CHANGE
		ws.ChangeServiceConfig(self.handle, self.snc, self.suptype, \
		self.snc, None, None, 0, None, None, None, self.lserv) 
		
	def getname(self):
		self.snames=ws.EnumServicesStatus(self.scmhandle)
		for i in self.snames:
			if i[0].lower() == self.userv.lower():
				return i[0], i[1]; break
			if i[1].lower() == self.userv.lower():
				return i[0], i[1]; break				
		print("Error: The", self.userv, "service doesn't seem to exist.")
		return None, None

		
if __name__ == "__main__":
	import sys
	
	try:
		ipaddr = sys.argv [1]
	except IndexError:
		print("IP Address is not provided")
		sys.exit (1)

	
			



import boto.ec2.cloudwatch
from boto.ec2.cloudwatch import MetricAlarm
import datetime

STATISTICS = ['Minimum', 'Maximum', 'Sum', 'Average', 'SampleCount']
UNITS = [
	'Seconds', 'Microseconds', 'Milliseconds', 'Bytes', 'Kilobytes', 'Megabytes', 'Gigabytes', 'Terabytes', 
	'Bits', 'Kilobits', 'Megabits', 'Gigabits', 'Terabits', 'Percent', 'Count', 'Bytes/Second', 
	'Kilobytes/Second', 'Megabytes/Second', 'Gigabytes/Second', 'Terabytes/Second', 'Bits/Second', 
	'Kilobits/Second', 'Megabits/Second', 'Gigabits/Second', 'Terabits/Second', 
	'Count/Second', None
]

PREF = {
	'CPUUtilization': ('Average', 'Percent'),
	'CPUCreditBalance': ('Average', 'Count'),
	'CPUCreditUsage': ('Average', 'Count'),		
	'DiskReadBytes': ('Average', 'Bytes'),
	'DiskReadOps': ('Sum', 'Bytes'),
	'DiskWriteBytes': ('Average', 'Bytes'),
	'DiskWriteOps': ('Sum', 'Bytes'),		
	'NetworkIn': ('Maximum', 'Bytes'),
	'NetworkOut': ('Maximum', 'Bytes'),
}

def create_alram (name, metric, comparison, threshold, dimensions, arn, period = 300, evaluation_periods = 2):
	return MetricAlarm(
		name = name, 
		metric = metric, 
		statistic = PREF [metric][0],
		comparison = comparison, 
		alarm_actions = [arn],
		dimensions = dimensions,
		namespace = 'AWS/EC2',
		threshold = str (threshold),
		period = str (period), 
		evaluation_periods = evaluation_periods
	)
	
def get_instance_metric (instance_id,region = 'us-east-1'):
	cw = boto.ec2.cloudwatch.connect_to_region(region)
	return cw.list_metrics(dimensions = {"InstanceId": instance_id})	
	
def get_latest (instance_id, measure = None):	
	end = datetime.datetime.utcnow()
	start = end - datetime.timedelta (minutes = 20)	
	r = {}
	for metric in get_instance_metric (instance_id):
		pref = measure or PREF.get (metric.name)
		if not pref:
			continue
		datapoints = metric.query (start, end, *pref)
		r [metric.name] = datapoints and datapoints [0][pref [0]] or None		
	return r		



import asyncore 
import re 
import time 
import sys 
import threading 	

DEBUG = False

class OperationalError (Exception):
	pass
	
class DBConnect:
	zombie_timeout = 120
	keep_alive = 120
	backend_keep_alive = 300
	
	def __init__ (self, address, params = None, lock = None, logger = None):
		self.address = address
		self.params = params		
		self.dbname, self.user, self.password = "", "", ""
		
		if self.params:
			self.dbname, auth = self.params
			if auth:			
				if type (auth) is not tuple:
					self.user = auth
				else:
					try: self.user, self.password = auth
					except ValueError: self.username = auth
						
		self.lock = lock
		self.logger = logger
		
		self.request_count = 0
		self.execute_count = 0
		
		self._cv = threading.Condition ()
		self.backend = False
		
		# need if there's not any request yet
		self.active = 0
		self.retried = 0
		self.request = None
		self.has_result = False
		self.__history = []
		self.__no_more_request = False
		
		self.set_event_time ()
	
	def set_backend (self, flag = True):
		self.backend = flag
		self.keep_alive = self.backend_keep_alive
		
	def duplicate (self):
		new_asyncon = self.__class__ (self.address, self.params, self.lock, self.logger)
		new_asyncon.keep_alive = self.keep_alive
		new_asyncon.backend = self.backend
		return new_asyncon
		
	def get_proto (self):
		# call by culster_manager
		return None
	
	def handle_abort (self):		
		# call by dist_call
		self.close ()
		
	def close (self, deactive = 1):
		self.request = None
		addr = type (self.address) is tuple and ("%s:%d" % self.address) or str (self.address)
		self.logger ("[info] .....dbo %s has been closed" % addr)	
		if deactive:
			self.set_active (False)
					
	def get_history (self):
		return self.__history
		
	def clean_shutdown_control (self, phase, time_in_this_phase):
		self.__no_more_request = True
		if self.isactive ():
			return 1
		else:
			self.handle_close ()
			self.__no_more_request = False
			return 0
	
	def empty_cursor (self):
		if self.has_result:
			self.fetchall ()			
	
	def maintern (self, object_timeout):
		# query done but not used
		if self.has_result and self.isactive () and time.time () - self.event_time > self.zombie_timeout:			
			self.empty_cursor ()
			self.set_active (False)
			
		if time.time () - self.event_time > object_timeout:
			if not self.isactive ():
				self.disconnect ()
				return True # deletable
		return False	
	
	def reconnect (self):
		self.disconnect ()
		self.connect ()
	
	def disconnect (self):
		# keep request, just close temporary
		request = self.request
		self.close (deactive = 0)
		self.request = request
	
	def close_case (self):
		raise NotImplementedError
			
	def set_active (self, flag, nolock = False):
		if not flag: 
			self.set_timeout (self.keep_alive)
			self.retried = 0
			
		if flag:
			flag = time.time ()
		else:
			flag = 0
			
		if nolock or self.lock is None:
			self.active = flag
			return
		
		self.lock.acquire ()
		self.active = flag
		self.request_count += 1		
		self.lock.release ()
	
	def get_active (self, nolock = False):
		if nolock or self.lock is None:
			return self.active			
		self.lock.acquire ()	
		active = self.active
		self.lock.release ()	
		return active
	
	def isactive (self):	
		return self.get_active () > 0
		
	def isconnected (self):	
		# self.connected should be defined at __init__ or asyncore
		return self.connected
		
	def get_request_count (self):	
		return self.request_count
	
	def get_execute_count (self):	
		return self.execute_count
	
	def connect (self, force = 0):
		raise NotImplementedError("must be implemented in subclass")
	
	def set_timeout (self, timeout):
		self.zombie_timeout = timeout

	def handle_timeout (self):
		self.handle_close (OperationalError, "Operation Timeout")
	
	def handle_error (self):
		dummy, exception_class, exception_str, tbinfo = asyncore.compact_traceback()
		self.has_result = False
		self.logger.trace ()
		self.handle_close (exception_class, exception_str)
	
	def handle_close (self, expt = None, msg = ""):
		if not expt and not self.retried:			
			self.disconnect ()
			self.retried = 1
			self.execute (self.request)			
			return
		
		if self.exception_class is None:	
			self.exception_class, self.exception_str = expt, msg		
		self.close_case_with_end_tran ()
		self.close ()
	
	def set_event_time (self):
		self.event_time = time.time ()			
	
	def log_history (self, msg):	
		self.__history.append ("BEGIN TRAN: %s" % sql)
	
	def close_case_with_end_tran (self):
		self.end_tran ()
		self.close_case ()
		
	#-----------------------------------------------------
	# DB methods
	#-----------------------------------------------------
	def fetchall (self):		
		raise NotImplementedError
	
	def end_tran (self):
		raise NotImplementedError
						
	def begin_tran (self, request):
		if self.__no_more_request:			
			try: 
				raise OperationalError ("Entered Shutdown Process")
			except: 
				self.handle_error ()
				return
		
		self.request = request	
		self.__history = []
		self.has_result = False
		self.exception_str = ""
		self.exception_class = None		
		self.execute_count += 1
		self.set_event_time ()		
		
	def execute (self, request):		
		self.begin_tran (request)		
		raise NotImplementedError("must be implemented in subclass")


class AsynDBConnect (DBConnect):
	def is_channel_in_map (self, map = None):
		if map is None:
			map = self._map
		return self._fileno in map
	
	def maintern (self, object_timeout):
		# when in map, mainteren by lifetime with zombie_timeout
		if self.is_channel_in_map ():
			return False
		return DBConnect.maintern (self, object_timeout)
		
	def del_channel (self, map=None):
		# do not remove self._fileno
		fd = self._fileno
		if map is None:
			map = self._map
		if fd in map:
			del map[fd]
	
	


import os, sys
from .processutil import is_running

# FILE LOCK

def mkdir (tdir, mod = -1):
	while tdir:
		if tdir [-1] in ("\\/"):
			tdir = tdir [:-1]
		else:
			break	

	if os.path.isdir (tdir): return	
	chain = [tdir]	
	while 1:
		tdir, last = os.path.split (tdir)		
		if not last: 
			break
		if tdir:
			chain.insert (0, tdir)
	
	for dir in chain [1:]:
		try: 
			os.mkdir (dir)
			if os.name == "posix" and mod != -1:
				os.chmod (dir, mod)				
		except OSError as why:
			if why.errno in (17, 183): continue
			else: raise


class PidFile:
	def __init__ (self, path):
		self.path = path
	
	def make (self):
		if self.isalive ():
			raise AssertionError("process is runnig. terminated.")
					
		pathtool.mkdir (self.path)
		pidfile = os.path.join (self.path, "pid")			
		f = open (pidfile, "w")
		f.write ("%s" % os.getpid ())
		f.close ()
	
	def remove (self):		
		pidfile = os.path.join (self.path, "pid")
		if os.path.isfile (pidfile):
			os.remove (pidfile)
		
	def getpid (self, match = None):
		pidfile = os.path.join (self.path, "pid")
		if os.path.isfile (pidfile):
			f = open (pidfile)
			pid = f.read ()
			f.close ()
			pid = int (pid)
			if is_running (pid, match):
				return pid
		return None
			
	def isalive (self):
		return self.getpid () is not None and True or False
				


# 2016. 1. 10 by Hans Roh hansroh@gmail.com

__version__ = "0.7.8.2"
version_info = tuple (map (lambda x: not x.isdigit () and x or int (x),  __version__.split (".")))
import threading
from . import lifetime, queue, request_builder, response_builder, stubproxy
from .lib import logger as logger_f
from .lib.athreads import trigger
from .client import socketpool
from .dbapi import dbpool
from .client import adns, asynconnect
from .lib.athreads.fifo import await_fifo
from . import client, dbapi
from aquests.protocols.dns.asyndns import async_dns
from .protocols.http import localstorage as ls
from .protocols.http import request_handler, response as http_response
from .protocols import http2
from .protocols.http2 import H2_PROTOCOLS
from .dbapi import request as dbo_request
import os
import asyncore
import timeit
import time, math, random
import sys

DEBUG = 0

try:
	from urllib.parse import urlparse
except ImportError:
	from urlparse import urlparse	

def cb_gateway_demo (response):
	global _logger 
	
	try: cl = len (response.content)
	except: cl = 0	
	if isinstance (response, dbo_request.Request):		
		status = "DBO %s %s %d records/documents received"	% (
			response.code, 
			response.msg,
			cl
		)
	else:
		status = "HTTP/%s %s %s %d bytes received" % (
			response.version,
			response.code, 
			response.msg, 
			cl
		)
			
	_logger.log (
		"REQ %s-%d. %s" % (
		response.meta ['req_method'],
		response.meta ['req_id'],		
		status
		)
	)	
	#print (response.headers)	
	#print (response.data)

	      		
_request_total = 0			
_finished_total = 0		
_initialized = False
_logger = None
_cb_gateway = cb_gateway_demo
_concurrent = 1
_workers = 1
_currents = {}
_que = None
_dns_query_req = {}
_timeout = 10
_max_conns = 0
_bytesrecv = 0
_allow_redirects = True
_force_h1 = False
	
def configure (
	workers = 1, 
	logger = None, 
	callback = None, 
	timeout = 10, 
	cookie = False, 
	force_http1 = False,
	http2_constreams = 1,
	allow_redirects = True,
	qrandom = False,
	use_pool = True,
	tracking = False,
	backend = False,
):
	global _logger, _cb_gateway, _concurrent, _initialized, _timeout
	global _workers, _que, _allow_redirects, _force_h1
	
	if logger is None:
		logger = logger_f.screen_logger ()
	_logger = logger
	
	if qrandom:		
		_que = queue.RandomQueue ()
	else:
		_que = queue.Queue ()
	
	_allow_redirects = allow_redirects
	_force_h1 = request_handler.RequestHandler.FORCE_HTTP_11 = force_http1
	
	if not use_pool:
		asynconnect.AsynConnect.keep_connect = use_pool
		asynconnect.AsynSSLConnect.keep_connect = use_pool		
	if not _force_h1:
		asynconnect.AsynConnect.fifo_class = await_fifo
		asynconnect.AsynSSLConnect.fifo_class = await_fifo
			
	http2.MAX_HTTP2_CONCURRENT_STREAMS = http2_constreams
	_workers = workers
	_concurrent = workers
	
	if not force_http1:
		_concurrent = workers * http2_constreams
	elif http2_constreams:
		_logger ("parameter http2_constreams is ignored", "warn")	
				
	if callback:
		_cb_gateway = callback
	
	if cookie:
		ls.create (_logger)
	_timeout = timeout	
	client.set_timeout (timeout)
	dbapi.set_timeout (timeout)
	
	socketpool.create (_logger, backend = backend, use_pool = use_pool)
	dbpool.create (_logger, backend = backend)
	adns.init (_logger)
	lifetime.init (_timeout / 2., logger) # maintern interval
	if tracking:
		lifetime.enable_memory_track ()
	_initialized = True

def _reque_first (request):
	global _que	
	_que.first (request)	

def handle_status_401 (response):
	global _que
	if not response.request.get_auth () or response.request.reauth_count:
		return response
	
	_logger ("authorization failed, %s" % response.url, "info")		
	request = response.request	
	request.reauth_count = 1	
	del response
	_reque_first (request)	

def handle_status_3xx (response):
	global _allow_redirects	, _que
	
	if not _allow_redirects:
		return response
	if response.status_code not in (301, 302, 307, 308):
		return response

	newloc = response.get_header ('location')	
	oldloc = response.request.uri
	request = response.request
	
	if newloc == oldloc:
		response.response = http_response.FailedResponse (711, "Redirect Error", request)
		return response
	
	try:		
		request.relocate (response.response, newloc)		
	except RuntimeError:		
		response.response = http_response.FailedResponse (711, "Redirect Error", request)
		return response
	
	_logger ("%s redirected to %s from %s" % (response.status_code, newloc, oldloc), "info")	
	# DO NOT use relocated response.request, it is None
	del response
	_reque_first (request)
	
def _request_finished (handler):
	global _cb_gateway, _currents, _concurrent, _finished_total, _logger, _bytesrecv,_force_h1
	try:
		if isinstance (handler, dbo_request.Request):
			response = handler
			_currents.pop (response.meta ['req_id'])
			
		else:
			response = response_builder.HTTPResponse (handler)
			_currents.pop (response.meta ['req_id'])
			for handle_func in (handle_status_401, handle_status_3xx):
				response = handle_func (response)
				if not response:
					return		
		_finished_total += 1
		response.logger = _logger
		_bytesrecv += len (response.content)
		callback = response.meta ['req_callback'] or _cb_gateway
		try:
			callback (response)
		except:		
			_logger.trace ()	
	
	except RecursionError:
		try: 
			_currents.pop (handler.request.meta ['req_id'])
		except KeyError: 
			pass	
		_logger ("too many error occured, stop requeueing", "info")	
		
	else:	
		qsize () and _req ()
	
	# clearing memory
	handler.request.meta = None
	del response
		
def _req ():
	global _que, _logger, _currents, _request_total, _backend
	args = _que.get ()
	
	if args is None:
		return
		
	_request_total += 1
	_is_request = False
	_is_db = False
	_method = None

	if type (args) is not tuple:
		req = args
		meta = req.meta
		_is_request = True
		_is_db = hasattr (req, 'dbtype')
		
	else:
		_is_request = False
		_method = args [0].lower ()
		
	if _is_db or _method in ("postgresql", "redis", "mongodb", "sqlite3"):
		if not _is_request:
			method, server, (dbmethod, params), dbname, auth, meta = args
			asyncon = dbpool.get (server, dbname, auth, "*" + _method)
			req = request_builder.make_dbo (_method, server, dbmethod, params, dbname, auth, meta, _logger)
		else:
			asyncon = dbpool.get (req.server, req.dbname, req.auth, "*" + req.dbtype)				
		_currents [meta ['req_id']] = [0, req.server]
		req.set_callback (_request_finished)
		asyncon.execute (req)
		
	else:
		if not _is_request:
			method, url, params, auth, headers, meta, proxy = args
			asyncon = socketpool.get (url)
			if _method in ("ws", "wss"):
				req = request_builder.make_ws (_method, url, params, auth, headers, meta, proxy, _logger)
			else:
				req = request_builder.make_http (_method, url, params, auth, headers, meta, proxy, _logger)			
		else:
			asyncon = socketpool.get (req.uri)		
		_currents [meta ['req_id']] = [0, req.uri]
		
		handler = req.handler (asyncon, req, _request_finished)
		if asyncon.get_proto () and asyncon.isconnected ():
			asyncon.handler.handle_request (handler)
		else:
			handler.handle_request ()	

def workings ():
	global _currents
	return 	len (_currents)
			
def countreq ():
	global _request_total
	return 	_request_total

def qsize ():	
	global _que
	return _que.qsize ()

def mapsize ():
	return len (asyncore.socket_map)
	
def countfin ():	
	global _finished_total
	return _finished_total

def countcli ():	
	global _currents
	return _currents

def concurrent ():
	global _concurrent
	return _concurrent

def fetchall ():
	global _workers, _logger, _que, _timeout, _max_conns, _bytesrecv, _concurrent, _finished_total, _max_conns, _force_h1	
	
	if not _initialized:
		configure ()
	
	_fetch_started = timeit.default_timer ()
	# IMP. mannually set
	lifetime._polling = 1
	
	# create initail workers	
	_logger ("creating connection pool", "info")
	target_socks = min (_workers, qsize ())
	for i in range (target_socks):
		_req ()
			
	if not _force_h1 and http2.MAX_HTTP2_CONCURRENT_STREAMS > 1:
		# wait all availabale	
		while qsize ():
			lifetime.lifetime_loop (os.name == "nt" and 1.0 or _timeout / 2.0, 1)
			if sum ([1 for conn in asyncore.socket_map.values () if not isinstance (conn, async_dns) and conn.get_proto () in H2_PROTOCOLS and conn.connected and not conn.isactive ()]) == _workers:
				_logger ('%d connection(s) created' % target_socks, 'info')
				break
			
	# now starting
	while qsize () or _currents:
		while _concurrent > min (len (_currents), mapsize ()) and qsize ():
			_req ()			
		_max_conns = max (_max_conns, mapsize ())	
		#print ('--', len (_currents), mapsize (), qsize ())
		if not mapsize ():
			break
		lifetime.lifetime_loop (os.name == "nt" and 1.0 or _timeout / 2.0, 1)
	
	#for each in _currents:
	#	print ('-- unfinished', each)
	
	lifetime._polling = 0	
	_duration = timeit.default_timer () - _fetch_started
	socketpool.cleanup ()
	dbpool.cleanup ()
	
	_logger.log ("* %d tasks during %1.2f sec (%1.2f tasks/s), recieved %d bytes (%d bytes/s), max %d conns" % (
		_finished_total, _duration, 
		_finished_total / _duration, 
		_bytesrecv,
		_bytesrecv / _duration,
		_max_conns
		)
	)

def suspend (timeout):
	a, b = math.modf (timeout)
	for i in range (int (b)):		
		socketpool.noop ()
		time.sleep (1)
	time.sleep (a)

_dns_reqs = 0
def _add (method, url, params = None, auth = None, headers = {}, callback = None, meta = None, proxy = None):	
	global _que, _initialized, _dns_query_req, _dns_reqs, _workers

	if not _initialized:		
		configure ()
		
	if not meta: 
		meta = {}

	meta ['req_id'] = _que.req_id
	meta ['req_method'] = method
	meta ['req_callback'] = callback
	host = urlparse (url) [1].split (":")[0]
	
	# DNS query for caching and massive
	if not lifetime._polling:
		if _dns_reqs < _workers and host not in _dns_query_req:
			_dns_query_req [host] = None
			_dns_reqs += 1
			adns.query (host, "A", callback = lambda x: None)		
		if mapsize ():
			lifetime.poll_fun_wrap (0.05)
		
	_que.add ((method, url, params, auth, headers, meta, proxy))
	
#----------------------------------------------------
# Add Reuqest (protocols.*.request) Object
#----------------------------------------------------	
def add (request):
	global _que	
	_que.add (request)
			
#----------------------------------------------------
# REST CALL
#----------------------------------------------------	
def get (*args, **karg):
	_add ('get', *args, **karg)

def delete (*args, **karg):
	_add ('delete', *args, **karg)

def head (*args, **karg):
	_add ('head', *args, **karg)

def trace (*args, **karg):
	_add ('trace', *args, **karg)

def options (*args, **karg):
	_add ('options', *args, **karg)
				
def post (*args, **karg):
	_add ('post', *args, **karg)

def patch (*args, **karg):
	_add ('patch', *args, **karg)

def patchform (*args, **karg):
	_add ('patchform', *args, **karg)

def patchxml (*args, **karg):
	_add ('patchxml', *args, **karg)

def patchjson (*args, **karg):
	_add ('patchjson', *args, **karg)

def patchnvp (*args, **karg):
	_add ('patchnvp', *args, **karg)
					
def put (*args, **karg):
	_add ('put', *args, **karg)

def postform (*args, **karg):
	_add ('postform', *args, **karg)

def postxml (*args, **karg):
	_add ('postxml', *args, **karg)

def postjson (*args, **karg):
	_add ('postjson', *args, **karg)	

def postnvp (*args, **karg):
	_add ('postnvp', *args, **karg)	

def putform (*args, **karg):
	_add ('putform', *args, **karg)

def putxml (*args, **karg):
	_add ('putxml', *args, **karg)

def putjson (*args, **karg):
	_add ('putjson', *args, **karg)	

def putnvp (*args, **karg):
	_add ('putnvp', *args, **karg)	
	
def upload (*args, **karg):
	_add ('upload', *args, **karg)	

def ws (*args, **karg):
	_add ('ws', *args, **karg)	

def wss (*args, **karg):
	_add ('wss', *args, **karg)	
	
	
#----------------------------------------------------
# XMLRPC, gRPC
#----------------------------------------------------
def _addrpc (method, rpcmethod, params, url, auth = None, headers = {}, callback = None, meta = {}, proxy = None):	
	_add (method, url, (rpcmethod, params), auth, headers, callback, meta, proxy)
	
def rpc	(*args, **karg):
	return stubproxy.Proxy ('rpc', _addrpc, *args, **karg)
	
def grpc	(*args, **karg):
	return stubproxy.Proxy ('grpc', _addrpc, *args, **karg)

#----------------------------------------------------
# DBO QEURY
#----------------------------------------------------
def _adddbo (method, dbmethod, params, server, dbname = None, auth = None, callback = None, meta = {}):
	global _que
	
	if not _initialized:
		configure ()
	if not meta: meta = {}	
	meta ['req_id'] = _que.req_id
	meta ['req_method'] = method
	meta ['req_callback'] = callback
	
	_que.add ((method, server, (dbmethod, params), dbname, auth, meta))
	
def postgresql (*args, **karg):
	return stubproxy.Proxy ('postgresql', _adddbo, *args, **karg)
pg = postgresql

def redis (*args, **karg):
	return stubproxy.Proxy ('redis', _adddbo, *args, **karg)

def mongodb (*args, **karg):
	return stubproxy.Proxy ('mongodb', _adddbo, *args, **karg)
	
def sqlite3 (*args, **karg):
	return stubproxy.Proxy ('sqlite3', _adddbo, *args, **karg)




import time
import boto.ec2
from boto.manage.cmdshell import sshclient_from_instance
from . import iquery

class SSH:
	def __init__ (self, instance_id, key_path):
		self.instance_id = instance_id
		self.instance = iquery.get_instance_by_id (instance_id)
		self.ssh_client = sshclient_from_instance(self.instance, key_path, user_name = "ubuntu")
		self.lines = []
	
	def __execute (self, cmd):	
		status, stdout, stderr = self.ssh_client.run (cmd)
		output = stderr.decode ('utf8') + stdout.decode ('utf8')
		if status != 0:
			raise SystemError (output)
		return output	
		
	def run (self):
		self.lines.insert (0, 'cat > .boto.sh << EOF\n#!/bin/bash')
		self.lines.append ('EOF')
		self.__execute ("\n".join (self.lines))
		self.lines = []
		return self.__execute ('chmod 755 .boto.sh && ./.boto.sh')
			
	def add (self, cmd):
		self.lines.append (cmd)

def get_shell_client (instance_id, key_path):
	return SSH (instance_id, key_path)
	


import collections

class AttrDict (dict):
	def __init__(self, *args, **kwargs):
		super(AttrDict, self).__init__(*args, **kwargs)
		self.__dict__ = self
	
class CaseInsensitiveKey(object):
	def __init__(self, key):
		self.key = key
		
	def __hash__(self):
		return hash(self.key.lower())
	
	def __eq__(self, other):
		return self.key.lower() == other.key.lower()
	
	def __str__(self):
		return self.key
	
	def __repr__(self):
		return self.key	
	
	def __getattr__ (self, name):
		return getattr (self.key, name)
		
# memory leaking?
class NocaseDict(dict):
	def __setitem__(self, key, value):
		key = CaseInsensitiveKey(key)
		super(NocaseDict, self).__setitem__(key, value)
	
	def __getitem__(self, key):
		key = CaseInsensitiveKey(key)
		return super(NocaseDict, self).__getitem__(key)


class CaseInsensitiveDict(collections.MutableMapping):
	def __init__(self, data=None, **kwargs):
		self._store = collections.OrderedDict()
		if data is None:
			data = {}
		self.update(data, **kwargs)

	def __setitem__(self, key, value):
		# Use the lowercased key for lookups, but store the actual
		# key alongside the value.
		self._store[key.lower()] = (key, value)

	def __getitem__(self, key):
		return self._store[key.lower()][1]

	def __delitem__(self, key):
		del self._store[key.lower()]

	def __iter__(self):
		return (casedkey for casedkey, mappedvalue in self._store.values())

	def __len__(self):
		return len(self._store)

	def lower_items(self):
		"""Like iteritems(), but with all lowercase keys."""
		return (
			(lowerkey, keyval[1])
			for (lowerkey, keyval)
			in self._store.items()
		)

	def __eq__(self, other):
		if isinstance(other, collections.Mapping):
			other = CaseInsensitiveDict(other)
		else:
			return NotImplemented
		# Compare insensitively
		return dict(self.lower_items()) == dict(other.lower_items())

	# Copy is required
	def copy(self):
		return CaseInsensitiveDict(self._store.values())

	def __repr__(self):
		return str(dict(self.items()))
				

if __name__ == "__main__":
	a = AttrDict ()
	a ["a-s"] = 1
	print (a.a_s)
	a.x = 4
	print (a['x'])
	
	b = NocaseDict ()
	b ['Content-Length'] = 100
	print (b ['content-length'])